#include <iostream>
#include <string>
#include <unistd.h>
#include <string.h>
#include <fstream>
#include <cvd/image_io.h>
#include <cvd/fast_corner.h>

using namespace CVD;
using namespace std;

#include "bibtex.cxx"

int main(int argc, char ** argv)
{
	cin.sync_with_stdio(false);
	cout.sync_with_stdio(false);
	
	bool list=0;
	int  threshold=20;
	int  n=9;
	bool nonmax=0;

	opterr=0;

	int i;
	while((i= getopt(argc, argv, "+lt:n:shb")) != -1)
	{
		string s;
		s += (char)i;
		if(i == 'l')	
			list = 1;
		else if (i == 't')
			threshold = atoi(optarg);
		else if (i == 'n')
		{
			n = atoi(optarg);
			if(n < 7 || n > 12)
			{
				cerr << "Error, n must be between 7 and 12.\n";
				return 1;
			}
				
		}
		else if (i == 's')
			nonmax = 1;
		else if (i == 'h')
		{
			cout << "fast: perform FAST feature detection on an image.\n"
					"\n"
				    "Edward Rosten\n"
					"\n"
				    "For more information on the FAST detector, visit the FAST home page at\n"
				    "http://mi.eng.cam.ac.uk/~er258/work/fast.html\n"
					"\n"
				    "If you use this in academic work, then please cite the following two papers:\n"
				    "    Fusing Points and Lines for High Performance Tracking, E. Rosten and T. Drummond, ICCV 2005\n"
				    "    Machine learning for high-speed corner detection, E. Rosten and T. Drummond, ECCV 2006\n"
				    "Rerun the program with -b to get the bibtex entries\n"
					"\n"
			        "invocation:\n"
					"\n"
				    "fast [ -hlnsbt ] [input_image [output_file]]\n"
					"\n"
				    "Options:\n"
					"\n"
				    "input_image        Input filename (default is stdin)\n"
				    "output_file        Output filename (default is stdout)\n"
				    "-l                 Output a list of corners (instead of an image)\n"
				    "-t X               Set threshold to X. Default is 20\n"
				    "-n X               Use X point fast (allowed range is 9--12, default is 9)\n"
				    "-s                 Perform nonmaximal suppression\n"
				    "-b                 Display bibtex entries\n"
				    "-h                 Display this message\n"
					"\n"
				    "Supported file types are: BMP, JPEG, TIFF (reading), PNM, EPS (writing).\n"
					"\n"
				    "example:\n"
					"\n"
				    "fast picture.jpg detected_corners.jpg\n";

			return 0;
		}
		else if ( i == 'b')
		{
			cout << bibtex;
			return 0;
		}
		else if ( i == '?')
		{
			cerr << "Error: -" << (char)optopt << " is not a valid option\n";
			return 1;
		}
	}
	
	try
	{
		ifstream inf;
		istream* in = &cin;
		ofstream outf;
		ostream* out=&cout;
		string out_name;
		
		if(optind < argc)
		{
			inf.open(argv[optind], ios::in|ios::binary);
			if(!inf.good())
				throw Exceptions::Image_IO::OpenError(argv[optind], "for reading", errno);

			in = &inf;
		}
		optind ++;

		if(optind < argc)
		{
			outf.open(argv[optind], ios::out|ios::binary);
			out_name = argv[optind];
			if(!outf.good())
				throw Exceptions::Image_IO::OpenError(argv[optind], "for writing", errno);

			out = &outf;
		}

		Image<byte> im;
		img_load(im, *in);

		vector<ImageRef> corners;
		
		if(n == 9)
			fast_corner_detect_9(im, corners, threshold);
		else if(n == 10)
			fast_corner_detect_10(im, corners, threshold);
		else if(n == 11)
			fast_corner_detect_11(im, corners, threshold);
		else if(n == 12)
			fast_corner_detect_12(im, corners, threshold);
		
		if(nonmax)
		{
			vector<ImageRef> c2;
			fast_nonmax(im, corners, threshold, c2);
			corners = c2;
		}

		if(list)
			for(unsigned int i=0; i < corners.size(); i++)
				(*out)  << corners[i].x << " " << corners[i].y << endl;
		else if(string_to_image_type(out_name) == ImageType::EPS || string_to_image_type(out_name) == ImageType::PS)
		{
				
			output_eps_header(*out, im);
			img_save(im, *out, ImageType::PS);

			*out << ".5 .5 translate\n";
			*out << "0.3 setlinewidth\n"	;
			*out << "/c{ 1 index 1 index 3 0 360 arc stroke .4 0 360 arc fill } bind def 1 0 0 setrgbcolor\n";
			*out << "1 0 0 setrgbcolor\n";

			for(unsigned int i=0; i < corners.size(); i++)
				*out << corners[i].x << " " << corners[i].y << " c\n";

			output_eps_footer(*out);
		}
		else
		{
			ImageRef cross[5] =
			{	
				ImageRef(0,0),
				ImageRef(1,0),
				ImageRef(0,1),
				ImageRef(-1,0),
				ImageRef(0,-1),
			};

			Rgb<byte> red(255, 0, 0);
			Image<Rgb<byte> > output = convert_image<Rgb<byte> >(im);
		
			for(unsigned int i=0; i < corners.size(); i++)
				for(int j=0; j < 5; j++)
					if(output.in_image(corners[i] + cross[j]))
						output[corners[i] + cross[j]] = red;
			
			img_save(output, *out, string_to_image_type(out_name));
		}
		

	}
	catch(Exceptions::All what)
	{
		cerr << "Error: " << what.what << endl;
		return 1;
	}

	return 0;

}
